@@ -26,6 +26,7 @@ from ._compat import IS_PY2
 from ._compat import IS_PY3
 from ._compat import IS_PY34_OR_GREATER
 from ._compat import IS_PY35_OR_GREATER
+from ._compat import IS_PY38_OR_GREATER
 
 import ast
 import contextlib
@@ -49,7 +50,9 @@ IOPERATOR_TO_STR = {
     ast.FloorDiv: '//='
 }
 
-if IS_PY35_OR_GREATER:
+if IS_PY2:  # pragma: PY2
+    pass  # trick branch coverage
+else:  # pragma: PY3
     IOPERATOR_TO_STR[ast.MatMult] = '@='
 
 
@@ -74,6 +77,32 @@ FORBIDDEN_FUNC_NAMES = frozenset([
     'breakpoint',
 ])
 
+# inspect attributes. See also
+# https://docs.python.org/3/library/inspect.html
+INSPECT_ATTRIBUTES = frozenset([
+    # traceback
+    "tb_frame",
+    "tb_next",
+    # code
+    "co_code",
+    # frame
+    "f_back",
+    "f_builtins",
+    "f_code",
+    "f_globals",
+    "f_locals",
+    "f_trace",
+    # generator
+    "gi_frame",
+    "gi_code",
+    "gi_yieldfrom",
+    # coroutine
+    "cr_await",
+    "cr_frame",
+    "cr_code",
+    "cr_origin",
+])
+
 
 # When new ast nodes are generated they have no 'lineno' and 'col_offset'.
 # This function copies these two fields from the incoming node
@@ -177,9 +206,9 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         return node
 
     def is_starred(self, ob):
-        if IS_PY3:
+        if IS_PY3:  # pragma: PY3
             return isinstance(ob, ast.Starred)
-        else:
+        else:  # pragma: PY2
             return False
 
     def gen_unpack_spec(self, tpl):
@@ -305,9 +334,9 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         try_body = [ast.Assign(targets=[target], value=converter)]
         finalbody = [self.gen_del_stmt(tmp_name)]
 
-        if IS_PY2:
+        if IS_PY2:  # pragma: PY2
             cleanup = ast.TryFinally(body=try_body, finalbody=finalbody)
-        else:
+        else:  # pragma: PY3
             cleanup = ast.Try(
                 body=try_body, finalbody=finalbody, handlers=[], orelse=[])
 
@@ -330,12 +359,12 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         return (tmp_target, cleanup)
 
     def gen_none_node(self):
-        if IS_PY34_OR_GREATER:
+        if IS_PY34_OR_GREATER:  # pragma: PY3
             return ast.NameConstant(value=None)
-        else:
+        else:  # pragma: PY2
             return ast.Name(id='None', ctx=ast.Load())
 
-    def gen_lambda(self, args, body):
+    def gen_lambda(self, args, body):  # pragma: PY2
         return ast.Lambda(
             args=ast.arguments(
                 args=args, vararg=None, kwarg=None, defaults=[]),
@@ -352,7 +381,11 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         Conversion is done by calling the 'slice' function from builtins
         """
 
-        if isinstance(slice_, ast.Index):
+        if isinstance(slice_, ast.expr):
+            # Python 3.9+
+            return slice_
+
+        elif isinstance(slice_, ast.Index):
             return slice_.value
 
         elif isinstance(slice_, ast.Slice):
@@ -422,7 +455,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         # which is gone in python3.
         # See https://www.python.org/dev/peps/pep-3113/
 
-        if IS_PY2:
+        if IS_PY2:  # pragma: PY2
             # Needed to handle nested 'tuple parameter unpacking'.
             # For example 'def foo((a, b, (c, (d, e)))): pass'
             to_check = list(node.args.args)
@@ -435,8 +468,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
 
             self.check_name(node, node.args.vararg)
             self.check_name(node, node.args.kwarg)
-
-        else:
+        else:  # pragma: PY3
             for arg in node.args.args:
                 self.check_name(node, arg.arg)
 
@@ -495,7 +527,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
             elif not print_used:
                 self.warn(node, "Doesn't print, but reads 'printed' variable.")
 
-    def gen_attr_check(self, node, attr_name):
+    def gen_attr_check(self, node, attr_name):  # pragma: PY2
         """Check if 'attr_name' is allowed on the object in node.
 
         It generates (_getattr_(node, attr_name) and node).
@@ -536,27 +568,65 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
 
     # ast for Literals
 
-    def visit_Num(self, node):
-        """Allow integer numbers without restrictions.
+    if IS_PY38_OR_GREATER:
 
-        Replaced by Constant in Python 3.8.
-        """
-        return self.node_contents_visit(node)
+        def visit_Constant(self, node):
+            """Allow constant literals with restriction for Ellipsis.
 
-    def visit_Str(self, node):
-        """Allow string literals without restrictions.
+            Constant replaces Num, Str, Bytes, NameConstant and Ellipsis in
+            Python 3.8+.
+            :see: https://docs.python.org/dev/whatsnew/3.8.html#deprecated
+            """
+            if node.value is Ellipsis:
+                # Deny using `...`.
+                # Special handling necessary as ``self.not_allowed(node)``
+                # would return the Error Message:
+                # 'Constant statements are not allowed.'
+                # which is only partial true.
+                self.error(node, 'Ellipsis statements are not allowed.')
+                return
+            return self.node_contents_visit(node)
 
-        Replaced by Constant in Python 3.8.
-        """
-        return self.node_contents_visit(node)
+    else:
 
-    def visit_Bytes(self, node):
-        """Allow bytes literals without restrictions.
+        def visit_Num(self, node):
+            """Allow integer numbers without restrictions.
 
-        Bytes is Python 3 only.
-        Replaced by Constant in Python 3.8.
-        """
-        return self.node_contents_visit(node)
+            Replaced by Constant in Python 3.8.
+            """
+            return self.node_contents_visit(node)
+
+        def visit_Str(self, node):
+            """Allow string literals without restrictions.
+
+            Replaced by Constant in Python 3.8.
+            """
+            return self.node_contents_visit(node)
+
+        def visit_Bytes(self, node):
+            """Allow bytes literals without restrictions.
+
+            Bytes is Python 3 only.
+            Replaced by Constant in Python 3.8.
+            """
+            return self.node_contents_visit(node)
+
+        def visit_Ellipsis(self, node):
+            """Deny using `...`.
+
+            Ellipsis exists only in Python 3.
+            Replaced by Constant in Python 3.8.
+            """
+            return self.not_allowed(node)
+
+        def visit_NameConstant(self, node):
+            """Allow constant literals (True, False, None) without ...
+
+            restrictions.
+
+            Replaced by Constant in Python 3.8.
+            """
+            return self.node_contents_visit(node)
 
     def visit_List(self, node):
         """Allow list literals without restrictions."""
@@ -574,28 +644,12 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """Allow dict literals without restrictions."""
         return self.node_contents_visit(node)
 
-    def visit_Constant(self, node):
-        """Allow constant literals without restrictions.
-
-        Constant replaces Num, Str, Bytes, NameConstant and Ellipsis in
-        Python 3.8+.
-        :see: https://docs.python.org/dev/whatsnew/3.8.html#deprecated
-        """
+    def visit_FormattedValue(self, node):
+        """Allow f-strings without restrictions."""
         return self.node_contents_visit(node)
 
-    def visit_Ellipsis(self, node):
-        """Allow using `...`.
-
-        Ellipsis is exists only in Python 3.
-        Replaced by Constant in Python 3.8.
-        """
-        return self.node_contents_visit(node)
-
-    def visit_NameConstant(self, node):
-        """Allow constant literals (True, False, None) without restrictions.
-
-        Replaced by Constant in Python 3.8.
-        """
+    def visit_JoinedStr(self, node):
+        """Allow joined string without restrictions."""
         return self.node_contents_visit(node)
 
     # ast for Variables
@@ -842,7 +896,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         # '*args' can be detected by 'ast.Starred' nodes.
         # '**kwargs' can be deteced by 'keyword' nodes with 'arg=None'.
 
-        if IS_PY35_OR_GREATER:
+        if IS_PY35_OR_GREATER:  # pragma: PY3
             for pos_arg in node.args:
                 if isinstance(pos_arg, ast.Starred):
                     needs_wrap = True
@@ -850,7 +904,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
             for keyword_arg in node.keywords:
                 if keyword_arg.arg is None:
                     needs_wrap = True
-        else:
+        else:  # pragma: PY2
             if (node.starargs is not None) or (node.kwargs is not None):
                 needs_wrap = True
 
@@ -895,6 +949,14 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
                 '"{name}" is an invalid attribute name because it ends '
                 'with "__roles__".'.format(name=node.attr))
 
+        if node.attr in INSPECT_ATTRIBUTES:
+            msg = ('"%s" is a restricted name,'
+                   ' that is forbidden to access in RestrictedPython.')
+            self.error(
+                node,
+                msg % node.attr,
+            )
+
         if isinstance(node.ctx, ast.Load):
             node = self.node_contents_visit(node)
             new_node = ast.Call(
@@ -1120,7 +1182,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
             raise NotImplementedError(
                 "Unknown target type: {0}".format(type(node.target)))
 
-    def visit_Print(self, node):
+    def visit_Print(self, node):  # pragma: PY2
         """Checks and mutates a print statement.
 
         Adds a target to all print statements.  'print foo' becomes
@@ -1178,7 +1240,7 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """Allow `as` statements in import and import from statements."""
         return self.node_contents_visit(node)
 
-    def visit_Exec(self, node):
+    def visit_Exec(self, node):  # pragma: PY2
         """Deny the usage of the exec statement.
 
         Exists only in Python 2.
@@ -1214,11 +1276,11 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """
         return self.node_contents_visit(node)
 
-    def visit_TryFinally(self, node):
+    def visit_TryFinally(self, node):  # pragma: PY2
         """Allow `try ... finally` without restrictions."""
         return self.node_contents_visit(node)
 
-    def visit_TryExcept(self, node):
+    def visit_TryExcept(self, node):  # pragma: PY2
         """Allow `try ... except` without restrictions."""
         return self.node_contents_visit(node)
 
@@ -1242,20 +1304,19 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """
         node = self.node_contents_visit(node)
 
-        if IS_PY3:
+        if IS_PY3:  # pragma: PY3
             self.check_name(node, node.name)
-            return node
+        else:   # pragma: PY2
+            if not isinstance(node.name, ast.Tuple):
+                return node
 
-        if not isinstance(node.name, ast.Tuple):
-            return node
+            tmp_target, unpack = self.gen_unpack_wrapper(node, node.name)
 
-        tmp_target, unpack = self.gen_unpack_wrapper(node, node.name)
+            # Replace the tuple with the temporary variable.
+            node.name = tmp_target
 
-        # Replace the tuple with the temporary variable.
-        node.name = tmp_target
-
-        # Insert the unpack code within the body of the except clause.
-        node.body.insert(0, unpack)
+            # Insert the unpack code within the body of the except clause.
+            node.body.insert(0, unpack)
 
         return node
 
@@ -1263,9 +1324,9 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """Protect tuple unpacking on with statements."""
         node = self.node_contents_visit(node)
 
-        if IS_PY2:
+        if IS_PY2:  # pragma: PY2
             items = [node]
-        else:
+        else:  # pragma: PY3
             items = node.items
 
         for item in reversed(items):
@@ -1294,24 +1355,22 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
             node = self.node_contents_visit(node)
             self.inject_print_collector(node)
 
-        if IS_PY3:
-            return node
+        if IS_PY2:  # pragma: PY2
+            # Protect 'tuple parameter unpacking' with '_getiter_'.
 
-        # Protect 'tuple parameter unpacking' with '_getiter_'.
+            unpacks = []
+            for index, arg in enumerate(list(node.args.args)):
+                if isinstance(arg, ast.Tuple):
+                    tmp_target, unpack = self.gen_unpack_wrapper(
+                        node, arg, 'param')
 
-        unpacks = []
-        for index, arg in enumerate(list(node.args.args)):
-            if isinstance(arg, ast.Tuple):
-                tmp_target, unpack = self.gen_unpack_wrapper(
-                    node, arg, 'param')
+                    # Replace the tuple with a single (temporary) parameter.
+                    node.args.args[index] = tmp_target
+                    unpacks.append(unpack)
 
-                # Replace the tuple with a single (temporary) parameter.
-                node.args.args[index] = tmp_target
-                unpacks.append(unpack)
-
-        # Add the unpacks at the front of the body.
-        # Keep the order, so that tuple one is unpacked first.
-        node.body[0:0] = unpacks
+            # Add the unpacks at the front of the body.
+            # Keep the order, so that tuple one is unpacked first.
+            node.body[0:0] = unpacks
         return node
 
     def visit_Lambda(self, node):
@@ -1320,46 +1379,46 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
 
         node = self.node_contents_visit(node)
 
-        if IS_PY3:
+        if IS_PY3:  # pragma: PY3
             # Implicit Tuple unpacking is not anymore available in Python3
             return node
+        else:  # pragma: PY2
+            # Check for tuple parameters which need _getiter_ protection
+            if not any(isinstance(arg, ast.Tuple) for arg in node.args.args):
+                return node
 
-        # Check for tuple parameters which need _getiter_ protection
-        if not any(isinstance(arg, ast.Tuple) for arg in node.args.args):
-            return node
+            # Wrap this lambda function with another. Via this wrapping it is
+            # possible to protect the 'tuple arguments' with _getiter_
+            outer_params = []
+            inner_args = []
 
-        # Wrap this lambda function with another. Via this wrapping it is
-        # possible to protect the 'tuple arguments' with _getiter_
-        outer_params = []
-        inner_args = []
+            for arg in node.args.args:
+                if isinstance(arg, ast.Tuple):
+                    tmp_name = self.gen_tmp_name()
+                    converter = self.protect_unpack_sequence(
+                        arg,
+                        ast.Name(tmp_name, ast.Load()))
 
-        for arg in node.args.args:
-            if isinstance(arg, ast.Tuple):
-                tmp_name = self.gen_tmp_name()
-                converter = self.protect_unpack_sequence(
-                    arg,
-                    ast.Name(tmp_name, ast.Load()))
+                    outer_params.append(ast.Name(tmp_name, ast.Param()))
+                    inner_args.append(converter)
 
-                outer_params.append(ast.Name(tmp_name, ast.Param()))
-                inner_args.append(converter)
+                else:
+                    outer_params.append(arg)
+                    inner_args.append(ast.Name(arg.id, ast.Load()))
 
-            else:
-                outer_params.append(arg)
-                inner_args.append(ast.Name(arg.id, ast.Load()))
+            body = ast.Call(func=node, args=inner_args, keywords=[])
+            new_node = self.gen_lambda(outer_params, body)
 
-        body = ast.Call(func=node, args=inner_args, keywords=[])
-        new_node = self.gen_lambda(outer_params, body)
+            if node.args.vararg:
+                new_node.args.vararg = node.args.vararg
+                body.starargs = ast.Name(node.args.vararg, ast.Load())
 
-        if node.args.vararg:
-            new_node.args.vararg = node.args.vararg
-            body.starargs = ast.Name(node.args.vararg, ast.Load())
+            if node.args.kwarg:
+                new_node.args.kwarg = node.args.kwarg
+                body.kwargs = ast.Name(node.args.kwarg, ast.Load())
 
-        if node.args.kwarg:
-            new_node.args.kwarg = node.args.kwarg
-            body.kwargs = ast.Name(node.args.kwarg, ast.Load())
-
-        copy_locations(new_node, node)
-        return new_node
+            copy_locations(new_node, node)
+            return new_node
 
     def visit_arguments(self, node):
         """
@@ -1400,9 +1459,9 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
         """Check the name of a class definition."""
         self.check_name(node, node.name)
         node = self.node_contents_visit(node)
-        if IS_PY2:
+        if IS_PY2:  # pragma: PY2
             new_class_node = node
-        else:
+        else:  # pragma: PY3
             if any(keyword.arg == 'metaclass' for keyword in node.keywords):
                 self.error(
                     node, 'The keyword argument "metaclass" is not allowed.')
@@ -1453,3 +1512,22 @@ class RestrictingNodeTransformer(ast.NodeTransformer):
     def visit_AsyncWith(self, node):
         """Deny async functionality."""
         self.not_allowed(node)
+
+    # Assignment expressions (walrus operator ``:=``)
+    # New in 3.8
+    def visit_NamedExpr(self, node):
+        """Allow assignment expressions under some circumstances."""
+        # while the grammar requires ``node.target`` to be a ``Name``
+        # the abstract syntax is more permissive and allows an ``expr``.
+        # We support only a ``Name``.
+        # This is safe as the expression can only add/modify local
+        # variables. While this may hide global variables, an
+        # (implicitly performed) name check guarantees (as usual)
+        # that no essential global variable is hidden.
+        node = self.node_contents_visit(node)  # this checks ``node.target``
+        target = node.target
+        if not isinstance(target, ast.Name):
+            self.error(
+                node,
+                "Assignment expressions are only allowed for simple targets")
+        return node
